/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Alkali (Basic) system.
 #
 # The purpose of this programme is to take advantage
 # of an assortment of hardware found around the
 # household.
 #
 # Hardware assumed to be attached to the ATmega8 are:
 #   - LEDs
 #   - Seven-Segment single-digit LED
 #   - Three pinned tachometer fan
 #   - 2 bit flip flop IC
 #
 #
 # Author: Eugene Kin Chee Yip
 # Date: 25 / 11 / 2008
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */


// AVR include files
#include <avr/io.h>
#include <avr/interrupt.h>

#include "System.h"


// Function prototypes
void initialise(void);
void startUp(void);
void wirelessAdapter(void);
void events(void);
void persistance(void);


// Global variables
time systemClock;			// System timer/clock
time latency;				// Latency for LED status
time wirelessTimer;			// Light sensing LED timer

io wirelessSensor;			// Wireless sensor at PIND4
io eventOutput;				// Event output at PIND0
io systemSeconds;			// LED blinking to seconds

memory flipFlops;			// 2 bit flip flop memory block

char digit;					// Single digit counter for the number of "short" light events

unsigned short int systemPeriod[1] = {0x7C};			// Period for one second
unsigned short int latencyPeriod[1] = {0x77};			// Period to keep event light on
unsigned short int wirelessPeriods[2] = {0x01, 0x20};	// Period to bias and poll the wireless/light sensor

char memoryLocations[4] = {'C', PINC0, 'C', PINC1};	// Input locations to the 2 bit memory block



/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Main entry point of the program.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
int main(void) {
	cli();					// Disable interrupts
	
	initialise();			// Initialise the system
	startUp();				// Start up the system
	
	sei();					// Enable interrupts


	for (; ; ) {						// System loop
		
		events();					// Counts the number of events and displays on the 7-seg
		wirelessAdapter();				// Keeps the wireless IO working
		persistance();					// Recongise and act on a persistant signal
		
		// Indicate system time
		if (systemClock.expired) {
			timerReset(&systemClock);
			ioSend(&systemSeconds, systemSeconds.isHigh ^ 0x01);
		}
	}
	
	return 0;
}



/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Initialise the system hardware.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
void initialise() {
	// Single digit counting
	digit = 0x00;
	
	// Seven segment display
	sevenSegOn();
	
	// Memory block
	memorySetLocations(&flipFlops, memoryLocations, 0x02);
	memoryReset(&flipFlops);

	// Timers
	timerStop(&systemClock);
	timerReset(&systemClock);
	timerSetPeriods(&systemClock, systemPeriod, 1);
	
	timerStop(&wirelessTimer);
	timerReset(&wirelessTimer);
	timerSetPeriods(&wirelessTimer, wirelessPeriods, 2);
	
	timerStop(&latency);
	timerReset(&latency);
	timerSetPeriods(&latency, latencyPeriod, 1);

	// Pins
	ioRegister(&wirelessSensor, 'D', PIND4);
	ioForward(&wirelessSensor);

	ioRegister(&eventOutput, 'D', PIND0);
	ioForward(&eventOutput);
	
	ioRegister(&systemSeconds, 'D', PIND7);
	ioForward(&systemSeconds);
	
	// PWM
	pwmReset('A');
	
	/* Timer0.  1/256 prescaling
	 * Enable overflow
	 ************************************/
	TCCR0 |= (1 << CS02);
	TIMSK |= (1 << TOIE0);
	
	/* External Interrupts for fan tachometer output.
	 ************************************/
	GICR |= (1 << INT1) | (1 << INT0);
	GIFR |= (1 << INTF1) | (1 << INTF0);
	MCUCR |= (1 << ISC11) | (1 << ISC10) | (1 << ISC01);
}


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Start up the system hardware.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
void startUp() {	
	// PWM
	pwmOn('A');

	// Timers
	timerStart(&systemClock);
	timerStart(&wirelessTimer);
	timerStart(&latency);
	
	// Seven segment display
	sevenSegClear();
	
	// Wireless / light sensor
	ioSend(&wirelessSensor, 0x01);
}


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Timer0 Overflow.  For ticking timers.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
ISR(TIMER0_OVF_vect) {
	// System clock (1 second)
	timerTick(&systemClock);
	timerExpire(&systemClock);
	
	// Wireless / light sensor (reverse bias <-> poll)
	timerTick(&wirelessTimer);
	timerExpire(&wirelessTimer);
	
	// Latency timer
	timerTick(&latency);
	timerExpire(&latency);
	
	// Measure period of signal
	timerTick(&wirelessSensor.timer);
}


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # INT0 event.  Tachometer output.  Trigger on falling
 # edge.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
ISR(INT0_vect) {
	sevenSegWrite(0xFF, 0x01);
}


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # INT1 Overflow.  Tachometer output.  Trigger on rising
 # edge.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
ISR(INT1_vect) {
	sevenSegWrite(0xFF, 0x00);
}




/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Respond to the light events at the wireless/light sensor.
 # Will write to the memory, write to the seven-segment LED
 # or set the fan PWM.
 #
 # Wireless timer should be reset after an event is
 # processed.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
void events() {
	switch (ioEventPresent(&wirelessSensor)) {
		case (0x01):	sevenSegWrite(digit, 0xFF);				// Refresh seven segment digit
						digit = (digit + 1) % 0x10;
						memoryWrite(&flipFlops, 0x00, 0x00);
						memoryWrite(&flipFlops, 0x01, 0x00);	// Clear the memory
						pwmSet('A', 0xFF);						// Turn fan off
						timerReset(&wirelessSensor.timer);
						break;
						
		case (0x02):	memoryWrite(&flipFlops, 0x01, 0x01);	// Write 0x10 to the memory
						pwmSet('A', 0x60);						// Turn fan speed to medium
						timerReset(&wirelessSensor.timer);
						break;
						
		case (0x03):	memoryWrite(&flipFlops, 0x00, 0x01);	// Write 0x01 to the memory
						pwmSet('A', 0x00);						// Turn fan speed to maximum
						timerReset(&wirelessSensor.timer);
						break;
						
		default:		break;
	}
}



/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Manages the wireless / light sensor so that it is
 # reversed bias or listening at the right time for the 
 # right period of time.
 #
 # Note that the level of the wireless pin is read just
 # before it is reversed biased.  This is acheived by
 # reading the pin in the reverse biasing cycle but as
 # the first instruction.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
void wirelessAdapter() {
	if (wirelessTimer.expired) {
		timerRotatePeriods(&wirelessTimer);			// Rotate the period between bias and polling
		if (wirelessTimer.periodIndex == 1) {		// Polling
			ioListen(&wirelessSensor);				//   - Tristate pin
			ioSend(&wirelessSensor, 0x00);			//   - Set pin low
		} else {									// Reverse Bias
			ioRefresh(&wirelessSensor);				//   - Read in value.  isHigh = 0 if light shone
			ioForward(&wirelessSensor);				//   - Set pin to output
			ioSend(&wirelessSensor, 0x01);			//   - Set pin high
		}
		
		timerReset(&wirelessTimer);					// Reset the timer after the expiration
	}
}


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Detects when a signal has persisted for a set amount of
 # time and asserts an LED.  After the signal disappears,
 # the LED will continue to be asserted for one "latency
 # period" until which point it will be deasserted.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */
void persistance() {
	if (wirelessSensor.timer.ticks > 0xA0) {
		ioSend(&eventOutput, 0x01);			// Assert output
		timerReset(&latency);				// Reset timing
		timerStart(&latency);				// Start timing
	} else {
		if (latency.expired) {
			ioSend(&eventOutput, 0x00);		// De-assert output
			timerStop(&latency);			// Stop timing
		}
	}
}


