/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Alkali (Basic) system library.
 #
 # The purpose of this library is to take advantage
 # of an assortment of hardware found around the
 # household.
 #
 # The suppored Hardware are:
 #   - Pins
 #   - Seven-Segment single-digit LED
 #   - flip flop IC
 #   - Timers
 #   - PWM
 #
 #
 # Author: Eugene Kin Chee Yip
 # Date: 25 / 11 / 2008
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */


// AVR include files
#include <avr/io.h>

#include "System.h"


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Timer.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Set an array of periods to use.
inline void timerSetPeriods(time *timer, unsigned short int *period, char periods) {
	(*timer).period = period;
	(*timer).periodIndex = 0;
	(*timer).periods = periods;
}

// "Start" the timer.  The timer is actually alive all the time as it adds the
// "on" state of the timer to the number of ticks.  But if "on" is zero, the
// number of ticks does not increase so the timer is "off".
inline void timerStart(time *timer) {
	(*timer).on = 1;
}

// "Stop" the timer.  Changes "on" to zero so that the number of ticks does not
// increase.
inline void timerStop(time *timer) {
	(*timer).on = 0;
}

// Allow the timer to tick by adding the value of "on" to the number of ticks.
inline void timerTick(time *timer) {
	(*timer).ticks += (*timer).on;
}

// Rotate the period being used by rotating the period index.  The period index
// is the array element being used.  The index will wrap around when it reaches
// the end of the array.
inline void timerRotatePeriods(time *timer) {
	(*timer).periodIndex = ++(*timer).periodIndex % (*timer).periods;
}

// Reset the timer.  This means setting the number of ticks to zero.  Expire
// the timer to reset the "expired" variable.  Note that the period index does
// not get set back to zero.
inline void timerReset(time *timer) {
	(*timer).ticks = 0;
	timerExpire(timer);
}

// If the number of ticks is greater than the period, the timer is considered to
// be expired.
inline void timerExpire(time *timer) {
	(*timer).expired = ((*timer).ticks > (*timer).period[(*timer).periodIndex]);
}




/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Seven Segment display.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Bit pattern for drawing the characters to the display.
const char digitPattern[18] =	{
									0b01111110,		// 0
									0b00011000,		// 1
									0b01101101,		// 2
									0b00111101,		// 3
									0b00011011,		// 4
									0b00110111,		// 5
									0b01110111,		// 6
									0b00011100,		// 7
									0b01111111,		// 8
									0b00011111,		// 9
									0b01011111,		// A
									0b01110011,		// b
									0b01100001,		// c
									0b01111001,		// d
									0b01101111,		// e
									0b01000111,		// F
									0b00000000,		// ' '		for clear / space
									0b01100111		// E		for error
								};

// Turn the display on by setting the pins to outputs.
void sevenSegOn() {
	DDRB |= (1 << DDB6) | (1 << DDB7);										// Pin 1 and 2
	DDRC |= (1 << DDC2) | (1 << DDC3) | (1 << DDC4) | (1 << DDC5);		// PIn 6, 7, 9 and 10
	DDRD |= (1 << DDD5) | (1 << DDD6);										// Pin 4 and 5
}

// Turn the display off by tristating the pins.
void sevenSegOff() {
	DDRB &= ~(1 << DDB6 | 1 << DDB7);										// Pin 1 and 2
	DDRC &= ~(1 << DDC2 | 1 << DDC3 | 1 << DDC4 | 1 << DDC5);				// PIn 6, 7, 9 and 10
	DDRD &= ~(1 << DDD5 | 1 << DDD6);										// Pin 4 and 5
}

// Clear the display by writing the ' ' character.
void sevenSegClear() {
	sevenSegWrite(0x10, 0x00);
}

// Write a character to the display.  The character is checked for special
// characters.  If it is a single hexidecimal digit, the value is used to
// reference the bit pattern array to determine which pins should be set.
void sevenSegWrite(char digit, char decimal) {
	if (digit != 0xFF) {												// Determine if it is a decimal update
		switch (digit > 0x0F) {
			case (1):	digit = (digit == 0x10) ? digit : 0x11;		// Determine if it is a clear request or overflow
						break;
			
			default:	sevenSeg.digit = digit;
						break;
		}
		
		sevenSeg.digit = digit;
		sevenSeg.decimal = decimal;
		
		const int pattern = digitPattern[(short int)digit];			// Get the required bit pattern
		
		// First 'and' the pins to set them low.
		PORTB &= 0b00111111;					// PINB6, Pin 1 & PINB7, Pin 2
		PORTD &= 0b11011111;					// PIND5, Pin 4
		PORTC &= 0b11000011;					// PINC2, Pin 6 & PINC3, Pin 7 & PINC4, Pin 9 & PINC5, Pin 10
		
		// Then 'or' the pins to set them high
		PORTB |= (pattern << 0 & 0b01000000) | (pattern << 2 & 0b10000000);		// PINB6, Pin 1 | PINB7, Pin 2
		PORTD |= (pattern << 1 & 0b00100000);										// PIND5, Pin 4
		PORTC |= (pattern >> 1 & 0b00000100) | (pattern << 1 & 0b00001000)			// PINC2, Pin 6 | PINC3, Pin 7
			  |  (pattern << 3 & 0b00010000) | (pattern << 5 & 0b00100000);		// PINC4, Pin 9 | PINC5, Pin 10
	}
	
	switch (decimal) {
		case (0x00):	PORTD &= ~(1 << PIND6);
						break;
		
		case (0x01):	PORTD |= (1 << PIND6);
						break;
					
		default:		break;
	}
}






/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Input / Output.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Registers a pin for input / output.  This entails assigning a port 
// letter and pin number.  A timer is included for abstracting signal
// information based on duration.
void ioRegister(io *input, const char port, const char pin) {
	(*input).port = port;
	(*input).pin = pin;
	(*input).edge = 1;				// Assume to be idle
	(*input).eventPresent = 0;		// Assume no event has happened
	
	timerStop(&(*input).timer);	// Stop timer
	timerReset(&(*input).timer);	// Reset timer
}

// Puts a pin into tristate mode.
void ioListen(io *input) {
	char mask = ~(1 << (*input).pin);
	
	switch ((*input).port) {
		case ('B'):	DDRB &= mask;
					break;
		
		case ('C'):	DDRC &= mask;
					break;
		
		case ('D'):	DDRD &= mask;
					break;
					
		default:	break;
	}
}

// Puts a pin into output mode.
void ioForward(io *input) {
	char mask = 1 << (*input).pin;

	switch ((*input).port) {
		case ('B'):	DDRB |= mask;
					break;
		
		case ('C'):	DDRC |= mask;
					break;
		
		case ('D'):	DDRD |= mask;
					break;
					
		default:	break;
	}
}

// Asserts the level of a pin.  The procedure is to clear the value
// of the pin and then assigns the value of "signal" to the pin.
void ioSend(io *input, char signal) {
	char maskOff = ~(1 << (*input).pin);
	char maskOn = signal << (*input).pin;

	switch ((*input).port) {
		case ('B'):	PORTB &= maskOff;
					PORTB |= maskOn;
					break;
		
		case ('C'):	PORTC &= maskOff;
					PORTC |= maskOn;
					break;
		
		case ('D'):	PORTD &= maskOff;
					PORTD |= maskOn;
					break;
					
		default:	break;
	}
	
	(*input).isHigh = signal != 0x00;
}

// Grabs the level of a pin.  Operates in active low mode. This means when the pin
// is idle, it will expect a high signal.
// Value of edge when idle is 1.
//                    active is -1.
// As the signal goes from active low to high (rising edge), an event is detected
// and stored.
void ioRefresh(io *input) {
	switch ((*input).port) {
		case ('B'):	(*input).isHigh = (PINB >> (*input).pin) & 0b01;
					break;
		
		case ('C'):	(*input).isHigh = (PINC >> (*input).pin) & 0b01;
					break;
		
		case ('D'):	(*input).isHigh = (PIND >> (*input).pin) & 0b01;
					break;
					
		default:	break;
	}

	// Detect edges in the signal.
	if ((*input).isHigh) {								// Idle
		if ((*input).edge == -1) {						//   If it was active but now idle
			(*input).edge = 1;							//   Positive edge
			
			if ((*input).timer.ticks < 0xA0) {			//   Abstract event based on duration
				(*input).eventPresent = 0x01;
			} else if ((*input).timer.ticks < 0xFF){
				(*input).eventPresent = 0x02;
			} else {
				(*input).eventPresent = 0x03;
			}
			
			(*input).timer.on = 0;						// Turn off the timer
		}
	} else {											// Active
		if ((*input).edge == 1) {						//   If it was idle and now active
			(*input).edge = -1;						//   Negative edge
			(*input).timer.on = 1;						//   Turn on the timer
		}
	}
}


// Get the event flag then set it to zero.
char ioEventPresent(io *input) {
	char event = (*input).eventPresent;
	if ( event > 0x00) {
		(*input).eventPresent = 0x00;
		return event;
	}
	return 0x00;
}




/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # PWM with Timer 1, channel A and B
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Enable the PWM output on channel A or B.
void pwmOn(char channel) {
	switch (channel) {
		case ('A'):	DDRB |= (1 << DDB1);
					break;
		
		case ('B'):	DDRB |= (1 << DDB2);
					break;
		
		default:	break;
	}
}

// Disable the PWM output on channel A or B.
void pwmOff(char channel) {
	switch (channel) {
		case ('A'):	DDRB &= ~(1 << DDB1);
					break;
		
		case ('B'):	DDRB &= ~(1 << DDB2);
					break;
		
		default:	break;
	}
}

// Set the duty cycle of the PWM.
void pwmSet(char channel, char offPeriod) {
	switch (channel) {
		case ('A'):	OCR1A = offPeriod;
					break;
		
		case ('B'):	OCR1B = offPeriod;
					break;
		
		default:	break;
	}
}

// Get the duty cycle value from channel A or B.
char pwmGet(char channel) {
	switch (channel) {
		case ('A'):	return OCR1A;
		
		case ('B'):	return OCR1B;
		
		default:	return 0xFF;
	}
}

// Reset the PWM register values.
void pwmReset(char channel) {
	TCCR1A |= (1 << COM1A1) | (1 << COM1A0) | (1 << COM1B1) | (1 << COM1B0) | (1 << WGM10);
	TCCR1B |= (1 << WGM12) | (1 << CS11) | (1 << CS10);
	pwmSet(channel, 0xFF);
}



/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # 2 bit memory (flip flop)
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Read the memory bit from the specified location.
// Currently this does not work as no read lines are specified
// by the memory struct.
char memoryRead(memory *block, char location) {
	char port = (*block).locations[location << 1];
	char pin = (*block).locations[(location << 1) + 1];

	switch (port) {
		case ('B'):	return (PINB >> pin) & 0b01;
					break;
		
		case ('C'):	return (PINC >> pin) & 0b01;
					break;
		
		case ('D'):	return (PIND >> pin) & 0b01;
					break;
					
		default:	return 0x00;
					break;
	}
}

// Write the memory bit to the specified location.  This requires a
// pulse signal to be generated for the write to happen.  The flip
// flop is performed with a synchronous "clock" signal.
void memoryWrite(memory *block, char location, char bit) {
	char port = (*block).locations[location << 1];
	char pin = (*block).locations[(location << 1) + 1];
	
	char maskOff = ~(1 << pin);
	char maskOn = (bit << pin);

	// Set the data line.
	switch (port) {
		case ('B'):	PORTB &= maskOff;
					PORTB |= maskOn;
					break;
		
		case ('C'):	PORTC &= maskOff;
					PORTC |= maskOn;
					break;
		
		case ('D'):	PORTD &= maskOff;
					PORTD |= maskOn;
					break;
					
		default:	break;
	}

	int i;
	for (i = 20000; i > 0; --i);
	
	// Pulse generation for committing the write.
	PORTB |= (1 << PINB0);
	for (i = 20000; i > 0; --i);

	PORTB &= ~(1 << PINB0);
	
	// Remove the data line.
	switch (port) {
		case ('B'):	PORTB &= maskOff;
					break;
		
		case ('C'):	PORTC &= maskOff;
					break;
		
		case ('D'):	PORTD &= maskOff;
					break;
					
		default:	break;
	}
	
}


// Set the location of memory bits based on an array.  The array will
// contain [PORT letter, PIN number, ..., ..., PORT letter, PIN number].
// The size of the memory block is equal to the number of memory bits.
void memorySetLocations(memory *block, char *locations, char size) {
	(*block).locations = locations;
	(*block).size = size;
}

// Reset the memory data to zero and also set the address lines to
// output mode.
void memoryReset(memory *block) {
	int i;
	for (i = 0; i < (*block).size; i++) {
		char port = (*block).locations[i << 1];
		char pin = (*block).locations[(i << 1) + 1];
		
		char mask = (1 << pin);
		
		switch (port) {
			case ('B'):	PORTB &= ~mask;
						DDRB |= mask;
						break;
			
			case ('C'):	PORTC &= ~mask;
						DDRC |= mask;
						break;
			
			case ('D'):	PORTD &= ~mask;
						DDRD |= mask;
						break;
						
			default:	break;
		}
	}
	for (i = 20000; i > 0; --i);
	
	// Pulse generation for committing the writes.
	DDRB |= (1 << PINB0);
	PORTB |= (1 << PINB0);
	for (i = 20000; i > 0; --i);

	PORTB &= ~(1 << PINB0);
}


