/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Alkali (Basic) system library.
 #
 # The purpose of this library is to take advantage
 # of an assortment of hardware found around the
 # household.
 #
 # The suppored Hardware are:
 #   - Pins
 #   - Seven-Segment single-digit LED
 #   - flip flop IC
 #   - Timers
 #   - PWM
 #
 #
 # Author: Eugene Kin Chee Yip
 # Date: 25 / 11 / 2008
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */



#ifndef	SYSTEM_UTILITIES
#define	SYSTEM_UTILITIES


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Timing functionality to hide the complexity of managing 
 # timer variables.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Timer struct
typedef struct {
	unsigned short int *period;			// Number of ticks before expiration.
	unsigned short int periodIndex;		// Current period in use
	char periods;						// Number of periods in period[]
	
	volatile unsigned long int ticks;	// Current number of ticks.
	volatile char expired;				// 0 if not expired, 1 if expired.
	char on;							// 0 if timer is off, 1 if timer is on.
} time;


// Timer functions
void timerExpire(time *);
void timerReset(time *);
void timerSetPeriods(time *, unsigned short int *, char);
void timerStart(time *);
void timerStop(time *);
void timerTick(time *);
void timerRotatePeriods(time *);



/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # 7-Seg display driver.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Hash table of pin patterns
const char digitPattern[18];

// 7-Seg struct
typedef struct {
	volatile char digit;		// Value on seven segment display.  0x10 indicates clear.  0x11 indicates an error
	volatile char decimal;		// 1 for on, 0 for off.
} sevenSegment;

sevenSegment sevenSeg;

// 7-Seg functions
void sevenSegOn(void);
void sevenSegOff(void);
void sevenSegClear(void);
void sevenSegWrite(char, char);


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # Input / Output driver.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// IO struct
typedef struct {
	char port;			// PINx,  Port letter
	char pin;				// PINxy, Pin number
	
	volatile char isHigh;		// Pin is high
	time timer;					// Timer for time context events
	
	short int edge;				// Store edge event of signal
	volatile char eventPresent;	// Indicates an unhandled event
} io;

// IO functions
void ioRegister(io *, char, char);
void ioListen(io *);
void ioForward(io *);
void ioSend(io *, char);
void ioRefresh(io *);
char ioEventPresent(io *);


/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # PWM driver for timer 1 and 2.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// PWM functions
void pwmOn(char);
void pwmOff(char);
void pwmSet(char, char);
char pwmGet(char);
void pwmReset(char);



/* # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
 #
 # 2 bit memory driver (2 Flip flops).
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # */

// Memory struct
typedef struct {
	char *locations;	// Location of memory bits
	char size;			// Size of memory in bits
} memory;

// Memory functions
char memoryRead(memory *, char);
void memoryWrite(memory *, char, char);
void memorySetLocations(memory *, char *, char);
void memoryReset(memory *);

#endif

